@@ -28,9 +28,7 @@
 package org.sat4j.minisat.constraints;
 
 import org.sat4j.minisat.constraints.card.AtLeast;
-import org.sat4j.minisat.constraints.cnf.LearntBinaryClause;
-import org.sat4j.minisat.constraints.cnf.LearntHTClause;
-import org.sat4j.minisat.constraints.cnf.UnitClause;
+import org.sat4j.minisat.constraints.cnf.LearntWLClause;
 import org.sat4j.minisat.core.Constr;
 import org.sat4j.specs.ContradictionException;
 import org.sat4j.specs.IVecInt;
@@ -44,13 +42,7 @@ public class CardinalityDataStructure extends AbstractCardinalityDataStructure {
     private static final long serialVersionUID = 1L;
 
     public Constr createUnregisteredClause(IVecInt literals) {
-    	if (literals.size()==1) {
-    		return new UnitClause(literals.last());
-    	}
-    	if (literals.size() == 2) {
-            return new LearntBinaryClause(literals,getVocabulary());
-        }
-        return new LearntHTClause(literals, getVocabulary());
+        return new LearntWLClause(literals, getVocabulary());
     }
 
     /*
