@@ -49,8 +49,8 @@
  *******************************************************************************/
 package org.sat4j.minisat.core;
 
-import static org.sat4j.core.LiteralsUtils.var;
 import static org.sat4j.core.LiteralsUtils.toDimacs;
+import static org.sat4j.core.LiteralsUtils.var;
 
 import java.io.PrintStream;
 import java.io.PrintWriter;
@@ -71,6 +71,8 @@ import org.sat4j.specs.ISolver;
 import org.sat4j.specs.IVec;
 import org.sat4j.specs.IVecInt;
 import org.sat4j.specs.IteratorInt;
+import org.sat4j.specs.Lbool;
+import org.sat4j.specs.SearchListener;
 import org.sat4j.specs.TimeoutException;
 
 /**
@@ -79,8 +81,8 @@ import org.sat4j.specs.TimeoutException;
  * 
  * @author leberre
  */
-public class Solver<L extends ILits, D extends DataStructureFactory<L>>
-		implements ISolver, UnitPropagationListener, ActivityListener, Learner {
+public class Solver<D extends DataStructureFactory> implements ISolver,
+		UnitPropagationListener, ActivityListener, Learner {
 
 	private static final long serialVersionUID = 1L;
 
@@ -138,15 +140,15 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 
 	private int[] model = null;
 
-	protected L voc;
+	protected ILits voc;
 
-	private IOrder<L> order;
+	private IOrder order;
 
 	private final ActivityComparator comparator = new ActivityComparator();
 
 	private final SolverStats stats = new SolverStats();
 
-	private final LearningStrategy<L, D> learner;
+	private final LearningStrategy<D> learner;
 
 	protected final AssertingClauseGenerator analyzer;
 
@@ -166,12 +168,12 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 
 	private RestartStrategy restarter;
 
-	private final Map<String, Integer> constrTypes = new HashMap<String, Integer>();
+	private final Map<String, Counter> constrTypes = new HashMap<String, Counter>();
 
 	private boolean isDBSimplificationAllowed = false;
 
-	private int learnedLiterals = 0;
-	
+	private final IVecInt learnedLiterals = new VecInt();
+
 	protected IVecInt dimacs2internal(IVecInt in) {
 		// if (voc.nVars() == 0) {
 		// throw new RuntimeException(
@@ -193,17 +195,16 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 	 * must be provided, else it won't work either.
 	 * 
 	 * @param acg
-	 * 		an asserting clause generator
+	 *            an asserting clause generator
 	 */
 
-	public Solver(AssertingClauseGenerator acg, LearningStrategy<L, D> learner,
-			D dsf, IOrder<L> order, RestartStrategy restarter) {
+	public Solver(AssertingClauseGenerator acg, LearningStrategy<D> learner,
+			D dsf, IOrder order, RestartStrategy restarter) {
 		this(acg, learner, dsf, new SearchParams(), order, restarter);
 	}
 
-	public Solver(AssertingClauseGenerator acg, LearningStrategy<L, D> learner,
-			D dsf, SearchParams params, IOrder<L> order,
-			RestartStrategy restarter) {
+	public Solver(AssertingClauseGenerator acg, LearningStrategy<D> learner,
+			D dsf, SearchParams params, IOrder order, RestartStrategy restarter) {
 		analyzer = acg;
 		this.learner = learner;
 		this.order = order;
@@ -217,7 +218,7 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 	 * heuristics must be changed prior to calling that method.
 	 * 
 	 * @param dsf
-	 * 		the internal factory
+	 *            the internal factory
 	 */
 	public final void setDataStructureFactory(D dsf) {
 		dsfactory = dsf;
@@ -227,6 +228,9 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 		order.setLits(voc);
 	}
 
+	/**
+	 * @since 2.1
+	 */
 	public void setSearchListener(SearchListener sl) {
 		slistener = sl;
 	}
@@ -263,7 +267,7 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 	}
 
 	public int nConstraints() {
-		return constrs.size()+trail.size()-learnedLiterals;
+		return constrs.size() + trail.size();
 	}
 
 	public void learn(Constr c) {
@@ -310,10 +314,31 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 					"Reference to the constraint to remove needed!"); //$NON-NLS-1$
 		}
 		Constr c = (Constr) co;
-		c.remove();
+		c.remove(this);
 		constrs.remove(c);
 		clearLearntClauses();
-		cancelLearntLiterals();
+		String type = c.getClass().getName();
+		constrTypes.get(type).dec();
+		return true;
+	}
+
+	/**
+	 * @since 2.1
+	 */
+	public boolean removeSubsumedConstr(IConstr co) {
+		if (co == null) {
+			throw new IllegalArgumentException(
+					"Reference to the constraint to remove needed!"); //$NON-NLS-1$
+		}
+		if (constrs.last() != co) {
+			throw new IllegalArgumentException(
+					"Can only remove latest added constraint!!!"); //$NON-NLS-1$
+		}
+		Constr c = (Constr) co;
+		c.remove(this);
+		constrs.pop();
+		String type = c.getClass().getName();
+		constrTypes.get(type).dec();
 		return true;
 	}
 
@@ -357,7 +382,7 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 			for (int i = 0; i < cs[type].size(); i++) {
 				if (cs[type].get(i).simplify()) {
 					// enleve les contraintes satisfaites de la base
-					cs[type].get(i).remove();
+					cs[type].get(i).remove(this);
 				} else {
 					cs[type].moveTo(j++, i);
 				}
@@ -386,7 +411,7 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 	 * Satisfait un litt?ral
 	 * 
 	 * @param p
-	 * 		le litt?ral
+	 *            le litt?ral
 	 * @return true si tout se passe bien, false si un conflit appara?t.
 	 */
 	public boolean enqueue(int p) {
@@ -397,11 +422,11 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 	 * Put the literal on the queue of assignments to be done.
 	 * 
 	 * @param p
-	 * 		the literal.
+	 *            the literal.
 	 * @param from
-	 * 		the reason to propagate that literal, else null
+	 *            the reason to propagate that literal, else null
 	 * @return true if the asignment can be made, false if a conflict is
-	 * 	detected.
+	 *         detected.
 	 */
 	public boolean enqueue(int p, Constr from) {
 		assert p > 1;
@@ -418,6 +443,9 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 		voc.setLevel(p, decisionLevel());
 		voc.setReason(p, from);
 		trail.push(p);
+		if (from != null) {
+			from.forwardActivity(claInc);
+		}
 		return true;
 	}
 
@@ -449,8 +477,7 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 			preason.clear();
 			assert confl != null;
 			confl.calcReason(p, preason);
-			if (confl.learnt())
-				claBumpActivity(confl);
+			learnedConstraintsDeletionStrategy.onConflictAnalysis(confl);
 			// Trace reason for p
 			for (int j = 0; j < preason.size(); j++) {
 				int q = preason.get(j);
@@ -483,7 +510,7 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 
 		Constr c = dsfactory.createUnregisteredClause(outLearnt);
 		slistener.learn(c);
-
+		learnedConstraintsDeletionStrategy.onConflict(c);
 		results.reason = c;
 
 		assert outBtlevel > -1;
@@ -546,12 +573,12 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 
 	/**
 	 * Setup the reason simplification strategy. By default, there is no reason
-	 * simplification. NOTE THAT REASON SIMPLIFICATION DOES NOT WORK WITH 
-	 * SPECIFIC DATA STRUCTURE FOR HANDLING BOTH BINARY AND TERNARY CLAUSES. 
+	 * simplification. NOTE THAT REASON SIMPLIFICATION DOES NOT WORK WITH
+	 * SPECIFIC DATA STRUCTURE FOR HANDLING BOTH BINARY AND TERNARY CLAUSES.
 	 * 
 	 * @param simp
-	 * 		the name of the simplifier (one of NO_SIMPLIFICATION,
-	 * 		SIMPLE_SIMPLIFICATION, EXPENSIVE_SIMPLIFICATION).
+	 *            the name of the simplifier (one of NO_SIMPLIFICATION,
+	 *            SIMPLE_SIMPLIFICATION, EXPENSIVE_SIMPLIFICATION).
 	 */
 	public void setSimplifier(String simp) {
 		Field f;
@@ -706,7 +733,7 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 	 * Propagate activity to a constraint
 	 * 
 	 * @param confl
-	 * 		a constraint
+	 *            a constraint
 	 */
 	public void claBumpActivity(Constr confl) {
 		confl.incActivity(claInc);
@@ -735,13 +762,13 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 		while (qhead < trail.size()) {
 			stats.propagations++;
 			int p = trail.get(qhead++);
-			slistener.propagating(toDimacs(p));
+			slistener.propagating(toDimacs(p), null);
 			order.assignLiteral(p);
 			// p is the literal to propagate
 			// Moved original MiniSAT code to dsfactory to avoid
 			// watches manipulation in counter Based clauses for instance.
 			assert p > 1;
-			IVec<Propagatable> watched = dsfactory.getAttachesFor(p);
+			IVec<Propagatable> watched = dsfactory.getWatchesFor(p);
 
 			final int size = watched.size();
 			for (int i = 0; i < size; i++) {
@@ -750,7 +777,7 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 					// Constraint is conflicting: copy remaining watches to
 					// watches[p]
 					// and return constraint
-					dsfactory.conflictDetectedInAttachesFor(p, i);
+					dsfactory.conflictDetectedInWatchesFor(p, i);
 					qhead = trail.size(); // propQ.clear();
 					// FIXME enlever le transtypage
 					return (Constr) watched.get(i);
@@ -765,7 +792,6 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 		slistener.adding(toDimacs(constr.get(0)));
 		if (constr.size() == 1) {
 			stats.learnedliterals++;
-			learnedLiterals++;
 		} else {
 			learner.learns(constr);
 		}
@@ -797,12 +823,15 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 	/**
 	 * Restore literals
 	 */
-	private void cancelLearntLiterals() {
+	private void cancelLearntLiterals(int learnedLiteralsLimit) {
+		learnedLiterals.clear();
 		// assert trail.size() == qhead || !undertimeout;
-		for (int c = learnedLiterals; c > 0; c--) {
+		while (trail.size() > learnedLiteralsLimit) {
+			learnedLiterals.push(trail.last());
 			undoOne();
 		}
 		qhead = trail.size();
+		// learnedLiterals = 0;
 	}
 
 	/**
@@ -877,7 +906,7 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 				// un conflit apparait
 				stats.conflicts++;
 				conflictC++;
-				slistener.conflictFound();
+				slistener.conflictFound(confl);
 				conflictCount.newConflict();
 				if (decisionLevel() == rootLevel) {
 					// on est a la racine, la formule est inconsistante
@@ -935,42 +964,18 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 		return fullmodel[var - 1];
 	}
 
-	/**
-     * 
-     */
-	protected void reduceDB() {
-		reduceDB(claInc / learnts.size());
-	}
-
 	public void clearLearntClauses() {
 		for (Iterator<Constr> iterator = learnts.iterator(); iterator.hasNext();)
-			iterator.next().remove();
+			iterator.next().remove(this);
 		learnts.clear();
+		learnedLiterals.clear();
 	}
 
-	protected void reduceDB(double lim) {
-		int i, j;
+	protected void reduceDB() {
 		sortOnActivity();
 		stats.reduceddb++;
-		for (i = j = 0; i < learnts.size() / 2; i++) {
-			Constr c = learnts.get(i);
-			if (c.locked()) {
-				learnts.set(j++, learnts.get(i));
-			} else {
-				c.remove();
-			}
-		}
-		for (; i < learnts.size(); i++) {
-			// Constr c = learnts.get(i);
-			// if (!c.locked() && (c.getActivity() < lim)) {
-			// c.remove();
-			// } else {
-			learnts.set(j++, learnts.get(i));
-			// }
-		}
-		System.out.println("c cleaning " + (learnts.size() - j) //$NON-NLS-1$
-				+ " clauses out of " + learnts.size() + " for limit " + lim); //$NON-NLS-1$ //$NON-NLS-2$
-		learnts.shrinkTo(j);
+		learnedConstraintsDeletionStrategy.reduce(learnts);
+		System.gc();
 	}
 
 	/**
@@ -1021,48 +1026,32 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 		return isSatisfiable(assumps, false);
 	}
 
-	public boolean isSatisfiable(IVecInt assumps, boolean global)
-			throws TimeoutException {
-		Lbool status = Lbool.UNDEFINED;
+	interface LearnedConstraintsDeletionStrategy extends Serializable {
 
-		final int howmany = voc.nVars();
-		if (mseen.length < howmany) {
-			mseen = new boolean[howmany + 1];
-		}
-		trail.ensure(howmany);
-		trailLim.ensure(howmany);
-		learnedLiterals = 0;
-		order.init();
-		learner.init();
-		restarter.init(params);
-		timebegin = System.currentTimeMillis();
-		slistener.start();
-		model = null; // forget about previous model
-		fullmodel = null;
+		void init();
 
-		// propagate constraints
-		Constr confl = propagate();
-		if (confl != null) {
-			analyzeAtRootLevel(confl);
-			slistener.end(Lbool.FALSE);
-			cancelUntil(0);
-			return false;
-		}
+		ConflictTimer getTimer();
 
-		// push incremental assumptions
-		for (IteratorInt iterator = assumps.iterator(); iterator.hasNext();) {
-			if (!assume(voc.getFromPool(iterator.next()))
-					|| (propagate() != null)) {
-				slistener.end(Lbool.FALSE);
-				cancelUntil(0);
-				return false;
-			}
-		}
-		rootLevel = decisionLevel();
+		void reduce(IVec<Constr> learnedConstrs);
+
+		void onConflict(Constr outLearnt);
+
+		void onConflictAnalysis(Constr reason);
+	}
+
+	/**
+	 * @since 2.1
+	 */
+	public LearnedConstraintsDeletionStrategy MEMORY_BASED = new LearnedConstraintsDeletionStrategy() {
+
+		/**
+		 * 
+		 */
+		private static final long serialVersionUID = 1L;
 
 		final long memorybound = Runtime.getRuntime().freeMemory() / 10;
 
-		ConflictTimer freeMem = new ConflictTimerAdapter(500) {
+		private final ConflictTimer freeMem = new ConflictTimerAdapter(500) {
 			private static final long serialVersionUID = 1L;
 
 			@Override
@@ -1076,6 +1065,204 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 			}
 		};
 
+		public void reduce(IVec<Constr> learnedConstrs) {
+			int i, j;
+			for (i = j = 0; i < learnts.size() / 2; i++) {
+				Constr c = learnts.get(i);
+				if (c.locked() || c.size() == 2) {
+					learnts.set(j++, learnts.get(i));
+				} else {
+					c.remove(Solver.this);
+				}
+			}
+			for (; i < learnts.size(); i++) {
+				learnts.set(j++, learnts.get(i));
+			}
+			System.out.println("c cleaning " + (learnts.size() - j) //$NON-NLS-1$
+					+ " clauses out of " + learnts.size()); //$NON-NLS-1$ //$NON-NLS-2$
+			learnts.shrinkTo(j);
+		}
+
+		public ConflictTimer getTimer() {
+			return freeMem;
+		}
+
+		@Override
+		public String toString() {
+			return "Memory based learned constraints deletion strategy";
+		}
+
+		public void init() {
+			// do nothing
+		}
+
+		public void onConflict(Constr outLearnt) {
+			// do nothing
+
+		}
+
+		public void onConflictAnalysis(Constr reason) {
+			if (reason.learnt())
+				claBumpActivity(reason);
+		}
+	};
+
+	/**
+	 * @since 2.1
+	 */
+	public LearnedConstraintsDeletionStrategy GLUCOSE = new LearnedConstraintsDeletionStrategy() {
+
+		/**
+		 * 
+		 */
+		private static final long serialVersionUID = 1L;
+		private int[] flags = new int[0];
+		private int flag = 0;
+
+		private final ConflictTimer clauseManagement = new ConflictTimerAdapter(
+				500) {
+			private static final long serialVersionUID = 1L;
+			private int nbconflict = 0;
+			private static final int MAX_CLAUSE = 20000;
+			private static final int INC_CLAUSE = 500;
+			private int nextbound = MAX_CLAUSE;
+
+			@Override
+			void run() {
+				nbconflict += INC_CLAUSE;
+
+				if (nbconflict >= nextbound) {
+					nextbound += INC_CLAUSE;
+					nbconflict = 0;
+					needToReduceDB = true;
+				}
+			}
+		};
+
+		public void reduce(IVec<Constr> learnedConstrs) {
+			int i, j;
+			for (i = j = learnedConstrs.size() / 2; i < learnedConstrs.size(); i++) {
+				Constr c = learnedConstrs.get(i);
+				if (c.locked() || c.getActivity() <= 2.0) {
+					learnedConstrs.set(j++, learnts.get(i));
+				} else {
+					c.remove(Solver.this);
+				}
+			}
+			System.out
+					.println("c cleaning " + (learnedConstrs.size() - j) //$NON-NLS-1$
+							+ " clauses out of " + learnedConstrs.size() + " with flag " + flag); //$NON-NLS-1$ //$NON-NLS-2$
+			learnts.shrinkTo(j);
+
+		}
+
+		public ConflictTimer getTimer() {
+			return clauseManagement;
+		}
+
+		@Override
+		public String toString() {
+			return "Glucose learned constraints deletion strategy";
+		}
+
+		public void init() {
+			final int howmany = voc.nVars();
+			if (flags.length <= howmany) {
+				flags = new int[howmany + 1];
+			}
+
+		}
+
+		public void onConflict(Constr outLearnt) {
+			// TODO Auto-generated method stub
+			int nblevel = 1;
+			flag++;
+			int currentLevel;
+			for (int i = 1; i < outLearnt.size(); i++) {
+				currentLevel = voc.getLevel(outLearnt.get(i));
+				if (flags[currentLevel] != flag) {
+					flags[currentLevel] = flag;
+					nblevel++;
+				}
+			}
+			outLearnt.incActivity(nblevel);
+		}
+
+		public void onConflictAnalysis(Constr reason) {
+			// do nothing
+		}
+	};
+
+	private LearnedConstraintsDeletionStrategy learnedConstraintsDeletionStrategy = MEMORY_BASED;
+
+	/**
+	 * @param lcds
+	 * @since 2.1
+	 */
+	public void setLearnedConstraintsDeletionStrategy(
+			LearnedConstraintsDeletionStrategy lcds) {
+		learnedConstraintsDeletionStrategy = lcds;
+	}
+
+	public boolean isSatisfiable(IVecInt assumps, boolean global)
+			throws TimeoutException {
+		Lbool status = Lbool.UNDEFINED;
+
+		final int howmany = voc.nVars();
+		if (mseen.length <= howmany) {
+			mseen = new boolean[howmany + 1];
+		}
+		trail.ensure(howmany);
+		trailLim.ensure(howmany);
+		learnedLiterals.ensure(howmany);
+		timebegin = System.currentTimeMillis();
+		slistener.start();
+		model = null; // forget about previous model
+		fullmodel = null;
+		order.init();
+		learnedConstraintsDeletionStrategy.init();
+
+		int learnedLiteralsLimit = trail.size();
+
+		// push previously learned literals
+		for (IteratorInt iterator = learnedLiterals.iterator(); iterator
+				.hasNext();) {
+			enqueue(iterator.next());
+		}
+
+		// propagate constraints
+		Constr confl = propagate();
+		if (confl != null) {
+			analyzeAtRootLevel(confl);
+			slistener.conflictFound(confl);
+			slistener.end(Lbool.FALSE);
+			cancelUntil(0);
+			cancelLearntLiterals(learnedLiteralsLimit);
+			return false;
+		}
+
+		// push incremental assumptions
+		for (IteratorInt iterator = assumps.iterator(); iterator.hasNext();) {
+			int p = voc.getFromPool(iterator.next());
+			if (!assume(p) || ((confl = propagate()) != null)) {
+				if (confl == null) {
+					slistener.conflictFound(p);
+				} else {
+					slistener.conflictFound(confl);
+				}
+				slistener.end(Lbool.FALSE);
+				cancelUntil(0);
+				cancelLearntLiterals(learnedLiteralsLimit);
+				return false;
+			}
+		}
+		rootLevel = decisionLevel();
+		// moved initialization here if new literals are added in the
+		// assumptions.
+		order.init(); // duplicated on purpose
+		learner.init();
+		restarter.init(params);
+
 		if (timeBasedTimeout) {
 			if (!global || timer == null) {
 				TimerTask stopMe = new TimerTask() {
@@ -1086,7 +1273,8 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 				};
 				timer = new Timer(true);
 				timer.schedule(stopMe, timeout);
-				conflictCount = freeMem;
+				conflictCount = learnedConstraintsDeletionStrategy.getTimer();
+				undertimeout = true;
 			}
 		} else {
 			if (!global || conflictCount == null) {
@@ -1099,13 +1287,13 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 						undertimeout = false;
 					}
 				};
+				undertimeout = true;
 				conflictCount = new ConflictTimerContainer().add(
-						conflictTimeout).add(freeMem);
+						conflictTimeout).add(
+						learnedConstraintsDeletionStrategy.getTimer());
 			}
 		}
 		needToReduceDB = false;
-		undertimeout = true;
-
 		// Solve
 		while ((status == Lbool.UNDEFINED) && undertimeout) {
 			status = search(restarter.nextRestartNumberOfConflict());
@@ -1116,13 +1304,16 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 		}
 
 		cancelUntil(0);
+		cancelLearntLiterals(learnedLiteralsLimit);
 		if (!global && timeBasedTimeout) {
 			timer.cancel();
 			timer = null;
 		}
 		slistener.end(status);
 		if (!undertimeout) {
-			throw new TimeoutException(" Timeout (" + timeout + "s) exceeded"); //$NON-NLS-1$//$NON-NLS-2$
+			String message = " Timeout (" + timeout
+					+ (timeBasedTimeout ? "s" : " conflicts") + ") exceeded";
+			throw new TimeoutException(message); //$NON-NLS-1$//$NON-NLS-2$
 		}
 		return status == Lbool.TRUE;
 	}
@@ -1130,7 +1321,28 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 	public void printInfos(PrintWriter out, String prefix) {
 		out.print(prefix);
 		out.println("constraints type ");
-		for (Map.Entry<String, Integer> entry : constrTypes.entrySet()) {
+		for (Map.Entry<String, Counter> entry : constrTypes.entrySet()) {
+			out.println(prefix + entry.getKey() + " => " + entry.getValue());
+		}
+	}
+
+	/**
+	 * @since 2.1
+	 */
+	public void printLearntClausesInfos(PrintWriter out, String prefix) {
+		Map<String, Counter> learntTypes = new HashMap<String, Counter>();
+		for (Iterator<Constr> it = learnts.iterator(); it.hasNext();) {
+			String type = it.next().getClass().getName();
+			Counter count = learntTypes.get(type);
+			if (count == null) {
+				learntTypes.put(type, new Counter());
+			} else {
+				count.inc();
+			}
+		}
+		out.print(prefix);
+		out.println("learnt constraints type ");
+		for (Map.Entry<String, Counter> entry : learntTypes.entrySet()) {
 			out.println(prefix + entry.getKey() + " => " + entry.getValue());
 		}
 	}
@@ -1139,25 +1351,29 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 		return stats;
 	}
 
-	public IOrder<L> getOrder() {
+	public IOrder getOrder() {
 		return order;
 	}
 
-	public void setOrder(IOrder<L> h) {
+	public void setOrder(IOrder h) {
 		order = h;
 		order.setLits(voc);
 	}
 
-	public L getVocabulary() {
+	public ILits getVocabulary() {
 		return voc;
 	}
 
 	public void reset() {
-		// FIXME verify that cleanup is OK
+		trail.clear();
+		trailLim.clear();
+		qhead = 0;
+		for (Iterator<Constr> iterator = constrs.iterator(); iterator.hasNext();)
+			iterator.next().remove(this);
+		constrs.clear();
+		clearLearntClauses();
 		voc.resetPool();
 		dsfactory.reset();
-		constrs.clear();
-		learnts.clear();
 		stats.reset();
 		constrTypes.clear();
 	}
@@ -1168,24 +1384,24 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 
 	/**
 	 * @param constr
-	 * 		a constraint implementing the Constr interface.
+	 *            a constraint implementing the Constr interface.
 	 * @return a reference to the constraint for external use.
 	 */
 	protected IConstr addConstr(Constr constr) {
 		if (constr != null) {
 			constrs.push(constr);
 			String type = constr.getClass().getName();
-			Integer count = constrTypes.get(type);
+			Counter count = constrTypes.get(type);
 			if (count == null) {
-				constrTypes.put(type, 1);
+				constrTypes.put(type, new Counter());
 			} else {
-				constrTypes.put(type, count + 1);
+				count.inc();
 			}
 		}
 		return constr;
 	}
 
-	public DataStructureFactory<L> getDSFactory() {
+	public DataStructureFactory getDSFactory() {
 		return dsfactory;
 	}
 
@@ -1197,7 +1413,7 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 	 * returns the ith constraint in the solver.
 	 * 
 	 * @param i
-	 * 		the constraint number (begins at 0)
+	 *            the constraint number (begins at 0)
 	 * @return the ith constraint
 	 */
 	public IConstr getIthConstr(int i) {
@@ -1221,6 +1437,7 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 				+ "speed (assignments/second)\t: " + stats.propagations //$NON-NLS-1$
 				/ cputime);
 		order.printStat(out, prefix);
+		printLearntClausesInfos(out, prefix);
 	}
 
 	/*
@@ -1231,7 +1448,7 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 	public String toString(String prefix) {
 		StringBuffer stb = new StringBuffer();
 		Object[] objs = { analyzer, dsfactory, learner, params, order,
-				simplifier, restarter };
+				simplifier, restarter, learnedConstraintsDeletionStrategy };
 		stb.append(prefix);
 		stb.append("--- Begin Solver configuration ---"); //$NON-NLS-1$
 		stb.append("\n"); //$NON-NLS-1$
@@ -1252,7 +1469,7 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 		stb.append(prefix);
 		stb.append("DB Simplification allowed=");
 		stb.append(isDBSimplificationAllowed);
-		stb.append("\n"); 
+		stb.append("\n");
 		stb.append(prefix);
 		stb.append("--- End Solver configuration ---"); //$NON-NLS-1$
 		return stb.toString();
@@ -1272,6 +1489,17 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 		return (int) (timeBasedTimeout ? timeout / 1000 : timeout);
 	}
 
+	/**
+	 * @since 2.1
+	 */
+	public long getTimeoutMs() {
+		if (!timeBasedTimeout) {
+			throw new UnsupportedOperationException(
+					"The timeout is given in number of conflicts!");
+		}
+		return timeout;
+	}
+
 	public void setExpectedNumberOfClauses(int nb) {
 		constrs.ensure(nb);
 	}
@@ -1306,6 +1534,33 @@ public class Solver<L extends ILits, D extends DataStructureFactory<L>>
 		isDBSimplificationAllowed = status;
 	}
 
+	/**
+	 * @since 2.1
+	 */
+	public int nextFreeVarId(boolean reserve) {
+		return voc.nextFreeVarId(reserve);
+	}
+
+	/**
+	 * @since 2.1
+	 */
+	public IConstr addBlockingClause(IVecInt literals)
+			throws ContradictionException {
+		return addClause(literals);
+	}
+
+	/**
+	 * @since 2.1
+	 */
+	public void unset(int p) {
+		int current = trail.last();
+		while (current != p) {
+			undoOne();
+			current = trail.last();
+		}
+		undoOne();
+	}
+
 }
 
 class ActivityComparator implements Comparator<Constr>, Serializable {
@@ -1318,7 +1573,11 @@ class ActivityComparator implements Comparator<Constr>, Serializable {
 	 * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
 	 */
 	public int compare(Constr c1, Constr c2) {
-		return (int) Math.round(c1.getActivity() - c2.getActivity());
+		long delta = Math.round(c1.getActivity() - c2.getActivity());
+		if (delta == 0) {
+			return c1.size() - c2.size();
+		}
+		return (int) delta;
 	}
 }
 
