@@ -8,20 +8,29 @@ import org.sat4j.specs.IVecInt;
 
 public class UnitClausePB extends UnitClause implements PBConstr {
 
-	public UnitClausePB(int value) {
+	private final ILits voc;
+
+	public UnitClausePB(int value, ILits voc) {
 		super(value);
+		this.voc = voc;
 	}
 
+	/**
+     * 
+     */
+	private static final long serialVersionUID = 1L;
+
 	public IVecInt computeAnImpliedClause() {
 		return null;
 	}
 
-	public BigInteger getCoef(int theLiteral1) {
+	public BigInteger getCoef(int literal) {
 		return BigInteger.ONE;
 	}
 
 	public BigInteger[] getCoefs() {
-		return new BigInteger[] {BigInteger.ONE};
+		BigInteger[] tmp = { BigInteger.ONE };
+		return tmp;
 	}
 
 	public BigInteger getDegree() {
@@ -29,11 +38,12 @@ public class UnitClausePB extends UnitClause implements PBConstr {
 	}
 
 	public int[] getLits() {
-		return new int[] {literal};
+		int[] tmp = { literal };
+		return tmp;
 	}
 
 	public ILits getVocabulary() {
-		return null;
+		return voc;
 	}
 
 }
